from .utility import Utility
class Segment:
    """  Represents a segment found in a Search.  """
    def __init__(self, indices, window, window_offset):
        self.indices = indices
        self.window = window
        self.window_offset = window_offset

class SearchResult:
    """ Represents a search result with information of chant and one or multiple found segments. """
    def __init__(self, chant, segments):
        self.chant = chant
        self.segments = segments

class Search:
    def __init__(self):
        pass

    @staticmethod
    def search_in_window(corpus, pattern, window, preprocess=None):
        """
        Searches for a pattern within a window in a given corpus.

        Args:
            corpus (Corpus): The corpus containing the documents to search.
            pattern (str): The pattern to search for.
            window (int): The window length for the search.
            preprocess (str): The window length for the search.

        Returns:
            list: List of search results containing segments and corresponding chant information.
        """
        search_result = []
        if preprocess == "intervals":
            data = [(document, Utility.compute_intervals(document.flat_neume_components)) for document in corpus.documents]
            for document in data:

                data_str = [f"{intv}" for intv in document[1]]

                segments = Search.search_pattern_in_window(data_str, pattern, window)
                if len(segments):
                    result_object = SearchResult(document[0], segments)
                    search_result.append(result_object)
        else:
            data = [(document, document.flat_neume_components) for document in corpus.documents]
            for document in data:

                data_str = [f"{note.base}{note.octave}" for note in document[1]]
                segments = Search.search_pattern_in_window(data_str, pattern, window)
                if len(segments):
                    result_object = SearchResult(document[0], segments)
                    search_result.append(result_object)
        return search_result

    @staticmethod
    def search_pattern_in_window(text, pattern, window_length):
        """
        Searches for a pattern within a window of text.

        Args:
            text (str): The text to search within.
            pattern (str): The pattern to search for.
            window_length (int): The window length for the search.

        Returns:
            list: List of segments containing window offset, indices, and the window.
        """
        text_length = len(text)
        segments = []
        for i in range(text_length - window_length + 1):
            window = text[i:i + window_length]
            window_indices = Search.is_pattern_in_window(window, pattern)
            if window_indices:
                segment = Segment(window_indices, window, i)
                segments.append(segment)
        return segments

    @staticmethod
    def is_pattern_in_window(window, pattern):
        """
        Checks if a pattern is present in the given window.

        Args:
            window (str): The window to search within.
            pattern (str): The pattern to search for.

        Returns:
            list: List of indices where the pattern is found in the window.
        """
        indices = []
        window_index = 0
        pattern_index = 0
        while window_index < len(window) and pattern_index < len(pattern):
            if window[window_index] == pattern[pattern_index]:
                indices.append(window_index)
                pattern_index += 1
            window_index += 1
        if pattern_index == len(pattern):
            return indices
        else:
            return None

    @staticmethod
    def visualize_html(results):
        """
        Generates an HTML representation of search results.

        Args:
            results (list): List of search results.

        Returns:
            str: HTML formatted representation of the search results.
        """
        html = "<html><head><style>.notation {font-family: Volpiano; font-size:2em;white-space:nowrap}</style></head><body>"
        html += (f"<h2>Results</h2>"
                f"<table>")

        for result in results:
            notation = ""
            notation += '<td class="notation">'
            close_span = False
            for index, pitch in enumerate(result.chant.volpiano):
                for segment in result.segments:
                    if close_span:
                        notation += "</span>"
                        close_span = False
                    if segment.window_offset + len(segment.window)-1 == index:
                        notation += '</td><td class="notation">'
                    if segment.window_offset == index:
                        notation += '</td><td class="notation">'

                    for id in segment.indices:
                        if (id + segment.window_offset) == index:
                            notation += '<span style="color: red">'
                            close_span = True
                notation += pitch
                notation += "-"
            notation += "</td>"
            html += (
                f'<tr><td>{result.chant.meta.initial_text} – {result.chant.meta.document_id} - {result.chant.meta.genre} </td>'
                f'{notation}</tr>')
        html += "</table></body></html>"
        return html

    @staticmethod
    def to_mafft_input(results, context=0):
        """
        Converts search results to input format for MAFFT.

        Args:
            results (list): List of search results.
            context (int): Additional context to include in the input.

        Returns:
            dict: Dictionary with formatted input for MAFFT.
        """
        mafft_input = {}
        for result in results:
            i = 1
            for segment in result.segments:
                i += 1
                name = f'{result.chant.meta.document_id.replace(" ", "")}-{i}'
                volp = result.chant.volpiano
                sl = slice(segment.window_offset - context, segment.window_offset + len(segment.window) + context)
                mafft_input[name] = "".join(volp[sl])
        return mafft_input



import unittest
from unittest.mock import MagicMock, patch
from monodikit.models.corpus import Corpus, create_document, check_files_exist
from monodikit.models.document import Chant, NeumeComponent, Syllable, Neume, Meta, Data
from monodikit.models.genre_specific import ProperTropeComplex
import tests.mock_data as mock

class TestNeume(unittest.TestCase):

    def test_calculate_number(self):
        neume = NeumeComponent("uuid", "C", False, "Normal", 4, False, (0,))
        self.assertEqual(neume.calculate_number(), 29)  # (4 * 7) + 3 = 29

    def test_parse_neume_content_flat(self):
        # Mock spaced_element for a flat note
        spaced_element = {
            "nonSpaced": [{"grouped": [{"uuid": "..",
                                        "focus": False,
                                        "noteType": "Flat",
                                        "base": "F", "octave": 3,
                                        "liquescent": False}]}],
        }

        neume = Neume(spaced_element, (0,))

        # Ensure neume has accidentals
        self.assertTrue(neume.accidentals)
        self.assertEqual(len(neume.accidentals), 1)

    def test_mei_generation(self):
        # Mock spaced_element for a normal note
        spaced_element = {
            "nonSpaced": [{"grouped": [{"uuid": "..",
                                        "focus": False,
                                        "noteType": "Normal",
                                        "base": "C", "octave": 4,
                                        "liquescent": False}]}],
        }

        neume = Neume(spaced_element, (0,))

        expected_mei = '<neume><nc pname="c" oct="4"/></neume>'
        self.assertEqual(neume.mei, expected_mei)


class TestSyllable(unittest.TestCase):
    def test_get_neumes(self):
        syllable = Syllable("uuid", "kind", (1,), "text", "Normal", mock.notes_data, False, False, True)
        self.assertEqual(len(syllable.neumes), 1)

    def test_mei_generation(self):
        syllable = Syllable("uuid", "kind", (0,), "text", "Normal", mock.notes_data, False, False, True, )
        expected_mei = '<syllable><syl>text</syl><neume><nc pname="c" oct="4"/></neume></syllable>'
        self.assertEqual(syllable.mei, expected_mei)


class TestVolpiano(unittest.TestCase):
    def test_one_doc(self):
        ground_truth = ['d', 'e', 'd', 'c', 'f', 'e', 'd', 'e', 'g', 'h', 'i', 'j', 'h', 'g', 'h', 'g', 'g', 'f', 'e', 'd', 'd', 'e', 'd',
         'c', 'f', 'g', 'f', 'e', 'd', 'e', 'g', 'h', 'i', 'j', 'h', 'g', 'f', 'e', 'd', 'e', 'f', 'e', 'd', 'c', 'd', 'h',
         'g', 'h', 'l', 'l', 'h', 'e', 'f', 'g', 'g', 'f', 'e', 'd', 'e', 'd', 'e', 'g', 'i', 'h', 'j', 'h', 'g', 'f', 'e',
         'd', 'c', 'e', 'f', 'e', 'd', 'c', 'd', 'd', 'f', 'h', 'd', 'e', 'd', 'c', 'g', 'e', 'd', 'l', 'l', 'h', 'k', 'j',
         'h', 'g', 'h', 'e', 'f', 'g', 'f', 'e', 'd', 'e', 'd', 'e', 'g', 'g', 'h', 'd', 'e', 'f', 'e', 'd', 'c', 'e', 'd',
         'e', 'f', 'e', 'd', 'c', 'd', 'd', 'f', 'g', 'h', 'i', 'j', 'h', 'g', 'f', 'e', 'd', 'f', 'e', 'c', 'd', 'f', 'g',
         'h', 'i', 'j', 'h', 'g', 'f', 'e', 'd', 'f', 'e', 'c', 'd', 'e', 'g', 'f', 'g', 'f', 'e', 'd', 'e', 'c', 'd', 'e',
         'f', 'g', 'f', 'e', 'd', 'e', 'd', 'e', 'g', 'f', 'g', 'e', 'd', 'e', 'c', 'd', 'e', 'f', 'g', 'f', 'e', 'd', 'e',
         'd', 'h', 'h', 'f', 'g', 'e', 'd', 'e', 'c', 'd', 'e', 'g', 'h', 'i', 'j', 'h', 'g', 'f', 'e', 'd', 'd', 'e', 'f',
         'e', 'd', 'c', 'd', 'h', 'h', 'f', 'g', 'e', 'd', 'e', 'c', 'd', 'e', 'g', 'h', 'i', 'j', 'h', 'g', 'f', 'e', 'd',
         'd', 'e', 'f', 'e', 'd', 'c', 'd', 'c', 'e', 'g', 'h', 'g', 'h', 'j', 'l', 'k', 'j', 'l', 'j', 'h', 'h', 'g', 'f',
         'e', 'd', 'e', 'd', 'd', 'e', 'f', 'e', 'd', 'c', 'e', 'f', 'g', 'f', 'e', 'd', 'e', 'd', 'c', 'e', 'g', 'h', 'g',
         'h', 'j', 'l', 'k', 'j', 'l', 'j', 'h', 'h', 'f', 'e', 'd', 'e', 'd', 'd', 'e', 'f', 'e', 'd', 'c', 'e', 'f', 'g',
         'f', 'e', 'd', 'e', 'd', 'c', 'e', 'g', 'h', 'g', 'h', 'j', 'k', 'j', 'h', 'l', 'j', 'h', 'h', 'd', 'h', 'k', 'j',
         'g', 'h', 'g', 'f', 'f', 'e', 'd', 'd', 'e', 'f', 'e', 'd', 'd', 'c', 'c', 'e', 'g', 'f', 'g', 'e', 'd', 'c', 'e',
         'g', 'h', 'g', 'h', 'j', 'k', 'j', 'h', 'l', 'j', 'h', 'h', 'd', 'h', 'h', 'k', 'j', 'g', 'h', 'g', 'f', 'f', 'e',
         'd', 'd', 'e', 'f', 'e', 'd', 'c', 'e', 'g', 'h', 'g', 'g', 'e', 'd', 'd', 'e', 'g', 'h', 'd', 'e', 'd', 'c', 'g',
         'f', 'e', 'd']
        doc = create_document(mock.chant_data["entry_path"])
        self.assertEqual(doc.volpiano, ground_truth)





import glob
import random
import os

from .document import Chant
from .source import create_source
import pickle


class Corpus:
    """
    A collection of chants loaded from a directory.

    Parameters:
    -----------
    directory : str
        The input directory that holds the data.

    sample : int, optional
        The number of chants that are given as a sample. Defaults to None.

    filter_options : dict or callable, optional
        If a dict is provided, it does a substring match.

        If a callable object is provided, it calls it for every document and creates the document
        if True is returned.

    Returns:
    --------
    Corpus
        A Corpus object containing the chants loaded from the directory.

    Examples:
    ---------
    Load chants with a specific substring match:
    ```
    subcorpus1 = Corpus("../data/*", filter_options={"gattung1": "Sequenz"})
    ```

    Load chants using a custom filtering function:
    ```
    subcorpus2 = Corpus("../data/*", filter_options=lambda chant: chant.gattung1 == "Tropus" or chant.gattung1 == "Sequenz")
    ```
    """

    def __init__(self, directory, sample=0, filters=None):
        if filters is None:
            filters = {}
        self.directory = directory
        self.filters = filters
        self.documents = []
        self.sample = False
        sources = [create_source(source) for source in glob.glob(self.directory)]
        self.sources = {source.sigle: source for source in sources}
        self.load_corpus(sample)

    # TODO: One should create subcorpora without the need to reload the whole corpus

    def __init__(self, directory=None, sample=0, filters=None, use_pkl=None):
        if use_pkl and os.path.exists(use_pkl):
            #print("pkl exists, use pkl..")
            with open(use_pkl, "rb") as f:
                corpus_data = pickle.load(f)
                self.directory = corpus_data.directory
                self.filters = corpus_data.filters
                self.documents = corpus_data.documents
                self.sample = corpus_data.sample
                self.sources = corpus_data.sources

        elif directory:
           # print("pkl does not exist.")

            self.directory = directory
            self.filters = filters
            self.documents = []
            self.sample = False
            sources = [create_source(source) for source in glob.glob(self.directory)]
            self.sources = {source.sigle: source for source in sources}
            self.load_corpus(sample)
            if use_pkl:
                print("save pkl.")
                with open(use_pkl, "wb") as f:
                    pickle.dump(self, f, pickle.HIGHEST_PROTOCOL)
        #print("end __init__")
            # TODO: One should create subcorpora without the need to reload the whole corpus

    def load_corpus(self, sample):
        if not sample:
            self.sample = False
            self.documents = list(filter(None, [create_document(entry,
                                                                filters=self.filters,
                                                                sources=self.sources)
                                                for source in glob.glob(self.directory)
                                                for entry in
                                                [directory for directory in
                                                 glob.glob(os.path.join(source, "*")) if
                                                 os.path.isdir(directory)]
                                                if check_files_exist(entry)]))  # ! Error when there are missing files?
        else:
            self.sample = True
            entries = [entry for source in glob.glob(self.directory)
                       for entry in
                       [directory for directory in glob.glob(os.path.join(source, "*"))
                        if os.path.isdir(directory)]]
            return list(filter(None, [create_document(entry, filters=self.filters)
                                      for entry in random.sample(entries, sample)
                                      if check_files_exist(entry)]))

    def filter(self, filters):
        if not callable(filters):
            Exception("Error  in filter(): argument has to be a callable")
        if self.filters:
            filters = lambda document_meta, source_meta: filters(document_meta, source_meta) and self.filters(document_meta, source_meta)
        filtered_corpus = Corpus()
        filtered_corpus.documents = [document for document in self.documents if filters(document.meta, self.sources[document.meta.source_id])]
        filtered_corpus.sources = self.sources
        filtered_corpus.filters = filters
        filtered_corpus.sample = self.sample
        return filtered_corpus



    @property
    def document_metadata(self):
        return [document.meta.as_record for document in self.documents]

    @property
    def source_metadata(self):
        return [source.meta.as_record for source in self.sources]


from .genre_specific import ProperTropeComplex


def create_document(entry, filters = None, sources=None):

    """
    Creates a new instance of a document.
    At first checks for a specific type and assigns a genre-specific subclass
    (for example TropeComplex()).
     If no suitable subclass is found, it returns a generic Document() instance.
     Example:
         ```
         document = create_document("./data/manuscriptid/documentid")
         ```
    """
    document_meta = Chant.get_meta(entry)
    if sources:
        source_meta = sources.get(document_meta.source_id, None)
    if filters:
        if callable(filters):
            if not filters(document_meta, source_meta):
                return None
        else:
            Exception("Filters have to be a callable")
   # else:
        # Now checks only for substring
       # for key, value in filters.items():
        #    if value not in document_meta.__getattribute__(key):
         #       return None

    if document_meta.genre == "Tropus":
        return ProperTropeComplex(entry)
    # ... other genres
    else:
        return Chant(entry)


def check_files_exist(path):
    meta_file = os.path.join(path, 'meta.json')
    data_file = os.path.join(path, 'data.json')

    if os.path.isfile(meta_file) and os.path.isfile(data_file):
        return True
    else:
        return False



import unittest
from unittest.mock import MagicMock, patch
from monodikit.models.corpus import Corpus, create_document, check_files_exist
from monodikit.models.document import Chant
from monodikit.models.genre_specific import ProperTropeComplex
import tests.mock_data as mock

class TestCorpus(unittest.TestCase):

    def test_create_document(self):
        # Mock data for the document creation
        entry = mock.chant_data["entry_path"]
        filters = lambda x, y: True  # Mock filter function
        sources = {"source_id": "Test"}  # Mock sources

        # Mock Chant class
        with patch('models.corpus.Chant', autospec=True) as mock_chant:
            result = create_document(entry, filters, sources)
            self.assertIsInstance(result, Chant)  # Check instance type


    # Add more test cases for the Corpus class functionality

if __name__ == '__main__':
    unittest.main()



import subprocess
class Synopsis:
    @staticmethod
    def run_mafft(input_data):
        """
        Runs the MAFFT multiple sequence alignment tool on the given input data.

        Args:
            input_data (dict): Dictionary with sequence names as keys and sequences as values.

        Returns:
            list: List of dictionaries containing sequence names and aligned sequences.
        """
        input_content = Synopsis.parse_mafft_input(input_data)
        with open("temp.fasta", "w") as f:
            f.write(input_content)
        command = f"mafft --auto --anysymbol temp.fasta"
        out = subprocess.run(command, shell=True, capture_output=True)

        return Synopsis.parse_mafft_output(str(out.stdout))

    @staticmethod
    def parse_mafft_input(input_data):
        """
        Parses the input data into the format expected by MAFFT.

        Args:
            input_data (dict): Dictionary with sequence names as keys and sequences as values.

        Returns:
            str: MAFFT-compatible formatted input.
        """
        output = ""
        for key in input_data:
            output += f"> {key}\n"
            output += f"{input_data[key]}\n"
        return output

    @staticmethod
    def parse_mafft_output(mafft_output):
        """
        Parses the MAFFT output into a list of aligned sequences with their names.

        Args:
            mafft_output (str): MAFFT output as a string.

        Returns:
            list: List of dictionaries containing sequence names and aligned sequences.
        """
        sequences = []
        blocks = mafft_output.strip().split('>')[1:]
        print("BLOCKS", blocks)

        for block in blocks:
            lines = block.strip().split('\\n')

            name = lines[0].strip().split()[0]
            sequence = ''.join(line.strip() for line in lines[1:])

            sequences.append({'name': name, 'sequence': sequence})
        return sequences

    @staticmethod
    def visualize_alignments(alignments, highlight=None, offset=0):
        """
        Generates an HTML table to visualize sequence alignments.

        Args:
            alignments (list): List of dictionaries containing sequence names and aligned sequences.
            highlight (list): List of lists specifying positions to highlight for each sequence. The result object of a search can be used.
            offset (int): Offset for highlighted positions.

        Returns:
            str: HTML formatted table for visualization.
        """
        html_output = "<table border='0'>"
        html_output += "<tr><th>Name</th><th>Sequence</th></tr>"
        for i, alignment in enumerate(alignments):
            name = alignment['name']
            sequence = ""
            if highlight:
                highlight_offset = [n + offset for n in highlight[i]]
                char_count = 0
                for char in alignment['sequence']:
                    if char == "-":
                        sequence += "-"
                    else:
                        if char_count in highlight_offset:
                            sequence += f"<span style='color: red'>{char}</span>"
                        else:
                            sequence += char
                        char_count += 1
            else:
                sequence = ",".join(list(alignment['sequence']))
            html_output += f'<tr><td>{name}</td><td style="text-align: left; font-family: Volpiano; font-size:3em; white-space: nowrap;">{sequence}</td></tr>'
        html_output += "</table>"
        return html_output



import glob, json


class Source:
    id: str
    sigle: str
    region_of_origin: str
    location_of_origin: str
    institution_of_origin: str
    religious_order: str
    type: str
    location_of_library: str
    library: str
    signature: str
    comment: str
    dating: str
    status: str
    century: str
    iiif_manifest_url: str
    iiif_foliooffset: str

    def __init__(self, id, quellensigle, herkunftsregion, herkunftsort, herkunftsinstitution, ordenstradition,
                 quellentyp,
                 bibliotheksort, bibliothek, bibliothekssignatur, kommentar, datierung, status, jahrhundert, manifest,
                 foliooffset,
                 ):
        self.id = id
        self.sigle = quellensigle
        self.region_of_origin = herkunftsregion
        self.location_of_origin = herkunftsort
        self.institution_of_origin = herkunftsinstitution
        self.religious_order = ordenstradition
        self.type = quellentyp
        self.location_of_library = bibliotheksort
        self.library = bibliothek
        self.signature = bibliothekssignatur
        self.comment = kommentar
        self.dating = datierung
        self.status = status
        self.century = jahrhundert
        self.iiif_manifest_url = manifest
        self.iiif_foliooffset = foliooffset

    def as_record(self):
        return {"id": self.id,
                "sigle": self.sigle,
                "region_of_origin": self.region_of_origin,
                "location_of_origin": self.location_of_origin,
                "institution_of_origin": self.institution_of_origin,
                "religious_order": self.religious_order,
                "type": self.type,
                "location_of_library": self.location_of_library,
                "library": self.library,
                "signature": self.signature,
                "comment": self.comment,
                "dating": self.dating,
                "status": self.status,
                "century": self.century,
                "iiif_manifest_url": self.iiif_manifest_url,
                "iiif_foliooffset": self.iiif_foliooffset
                }


def create_source(path):
    if glob.glob(path + "/meta.json"):
        with open(path + "/meta.json") as f:
            metadata = json.load(f)
            return Source(**metadata)
    else:
        raise Exception(f"Could not load source meta file {path}/meta.json")



class Utility:
    @staticmethod
    def compute_intervals(neume_components, start_asterisk = False, filter_division_change = False):
        note_to_num = {'C': 1, 'D': 2, 'E': 3, 'F': 4, 'G': 5, 'A': 6, 'B': 7}
        if start_asterisk:
            intervals = ["*"]
        else:
            intervals = []
        count_div_changes = 0
        for i in range(1, len(neume_components)):
            if filter_division_change and neume_components[i-1].index[0:-4] != neume_components[i].index[0:-4]:
                count_div_changes += 1
                continue
            prev_note_val = (neume_components[i - 1].octave * 7) + note_to_num[neume_components[i - 1].base]
            curr_note_val = (neume_components[i].octave * 7) + note_to_num[neume_components[i].base]
            interval = curr_note_val - prev_note_val
            if interval >= 0:
                intervals.append(interval)
            else:
                intervals.append(interval)
        return intervals

    @staticmethod
    def volpiano_view(volpiano_code, delimiter=",", size="20pt"):
        return f'<span style="font-family: volpiano; font-size: {size};">{delimiter.join(volpiano_code)}</span>'


"""
.. include:: ../Readme.md
.. include:: ../Examples.md
"""
from .models.document import (
    Chant,
    Division,
    Syllable,
    Neume,
    NeumeComponent
)
from .models.corpus import (
    Corpus
)
from .models.genre_specific import (
    ProperTropeComplex,
    Song,
    Sequence,
    PlayPassage,
    Play,
    OrdinaryChant
)
from .analysis.search import (
    Search
)
from .analysis.utility import (
    Utility,
)
from .analysis.synopsis import (
    Synopsis
)



import glob
import json
from dataclasses import dataclass, field
from typing import List
from xml.sax.saxutils import escape
import functools


@dataclass()
class Accidental:
    uuid: str
    base: str
    liquescent: str
    noteType: str
    octave: int
    focus: bool
    index: tuple

    @property
    def mei(self):
        if self.noteType == "Flat":
            accid = "f"
        elif self.noteType == "Natural":
            accid = "n"
        else:
            accid = "?"

        return f'<accid ploc="{self.base}" poct="{self.oct}" accid="{accid}"'

    @property
    def volpiano(self):
        if self.noteType == "Flat":
            if self.base == "B":
                return "i"
            if self.base == "E" and self.octave == 4:
                return "w"
            if self.base == "E" and self.octave == 5:
                return "x"
        elif self.noteType == "Natural":
            if self.base == "B":
                return "I"
            if self.base == "E" and self.octave == 4:
                return "W"
            if self.base == "E" and self.octave == 5:
                return "X"

    @property
    def json(self):
        return {"type": "accidental", "pitch": f"{self.base}{self.octave}"}


@dataclass
class NeumeComponent:
    uuid: str
    base: str
    liquescent: bool
    noteType: str
    octave: int
    focus: bool
    index: tuple
    note_to_num = {'C': 1, 'D': 2, 'E': 3, 'F': 4, 'G': 5, 'A': 6, 'B': 7}
    volpiano_matching = {"F3": "8", "G3": "9", "A3": "a", "B3": "b", "C4": "c", "D4": "d", "E4": "e", "F4": "f",
                         "G4": "g", "A4": "h", "B4": "j", "C5": "k", "D5": "l", "E5": "m", "F5": "n", "G5": "o",
                         "A5": "p", "B5": "q", "C6": "r", "D6": "s"}

    def calculate_number(self):
        return (self.octave * 7) + (self.note_to_num[self.base])

    def __lt__(self, other):
        return self.calculate_number() < other.calculate_number()

    def __gt__(self, other):
        return self.calculate_number() > other.calculate_number()

    def __le__(self, other):
        return self.calculate_number() <= other.calculate_number()

    def __ge__(self, other):
        return self.calculate_number() >= other.calculate_number()

    def __eq__(self, other):
        return self.calculate_number() == other.calculate_number() and self.liquescent == other.liquescent

    @property
    def pitch(self):
        return self.base + str(self.octave)

    @property
    def volpiano(self):
        return self.volpiano_matching[self.pitch]

    @property
    def mei(self):
        if self.index[-1] == 0:
            con = ""
        else:
            con = ' con="g"'

        if self.liquescent:
            return f'<nc pname="{self.base.lower()}" oct="{self.octave}"{con}><liquescent/></nc>'
        elif self.noteType != "Normal":
            return f'<nc pname="{self.base.lower()}" oct="{self.octave}"{con}><{self.noteType}/></nc>'
        else:
            return f'<nc pname="{self.base.lower()}" oct="{self.octave}"{con}/>'

    @property
    def json(self):
        return {"type": "note", "pitch": f"{self.base}{self.octave}"}

    @property
    def pitch(self):
        return f"{self.base}{self.octave}"

    def __str__(self):
        return f"<NeumeComponent base={self.base}, oct={self.octave}>"

    def __repr__(self):
        return f"<NeumeComponent base={self.base}, oct={self.octave}>"


@dataclass
class EmptyNeumeComponent(NeumeComponent):
    @property
    def pitch(self):
        return "Empty"

    @property
    def mei(self):
        return ""


class Neume:
    """
    A class representing a neume loosely following the MEI specification.

    """

    def __init__(self, spaced_element, index):  #: Takes a dictionary representing the spaced element of the neume.
        self.index = index
        self.neume_content = self.get_neume_content(spaced_element)  #: A list of `NeumeComponent` objects.
        self.neume_components = [element for element in self.neume_content if type(element) == NeumeComponent]
        self.accidentals = [element for element in self.neume_content if type(element) == Accidental]

    """Parses NeumeComponents and Accidentals within a Neume object."""
    def parse_neume_content(self, element, index):
        try:
            if element["noteType"] == "Normal":
                neume = NeumeComponent(**element, index=self.index + (index,))
                low_peak = NeumeComponent(uuid="", base="G", liquescent=False, noteType="Normal", octave=3, focus=False,
                                          index=self.index + (0,))
                comment_note = NeumeComponent(uuid="", base="A", liquescent=True, noteType="Normal", octave=5,
                                              focus=False, index=self.index + (0,))
                if neume < low_peak or neume == comment_note:
                    # return EmptyNeumeComponent(**element, index=self.index + (index,))
                    return None
                else:
                    return neume
            elif element["noteType"] == "Flat" or element.noteType == "Natural":
                return Accidental(**element, index=self.index + (index,))
        except AttributeError:
            return None

    """Wraps whole content up into a list"""
    def get_neume_content(self, spaced_element):
        return [neume for neume in [self.parse_neume_content(connected_neume_component, (index1 + index2))
                                    for index2, neume_component in enumerate(spaced_element["nonSpaced"])
                                    for index1, connected_neume_component in enumerate(neume_component["grouped"])] if
                neume is not None]

    @property
    def mei(self):
        neume_components = "".join([nc.mei for nc in self.neume_components])
        if len(self.accidentals):
            accidentals = "".join([accid.mei for accid in self.accidentals])
        else:
            accidentals = ""
        return f"{accidentals}<neume>{neume_components}</neume>"

    @property
    def json(self):
        return {"type": "neume", "elements": [neume_components.json for neume_components in self.neume_content]}


@dataclass
class Syllable:
    uuid: str
    kind: str
    index: tuple
    text: str = ""
    syllableType: str = ""
    notes: dict = field(default_factory=dict)  # Notes good terminology? includes groups etc.
    endsWord: bool = False
    focus: bool = False
    hasNotes: bool = None
    neumes: list = field(init=False)  # How to name elements / notes? notes is now input, elements the

    # processed Notes instance
    def __post_init__(self):
        if self.syllableType == "Normal":
            self.neumes = self.get_neumes(self.notes)
        else:
            self.neumes = []

    def get_neumes(self, notes):
        if "spaced" not in notes:
            return []
        return [Neume(neume, self.index + (index,)) for index, neume in enumerate(notes["spaced"])]

    @property
    def mei(self):
        neumes = "".join([neume.mei for neume in self.neumes])
        return f"<syllable><syl>{escape(self.text)}</syl>{neumes}</syllable>"

    @property
    def json(self):
        return {"type": "syllable", "lyric": self.text, "elements": [neume.json for neume in self.neumes]}



@dataclass
class EditorialLine:
    """
    A class representing a Editorial Line associated with the textual phrases.
    """
    uuid: str
    kind: str
    children: list
    index: tuple

    def __post_init__(self):
        self.syllables = self.get_syllables()

    def get_syllables(self):
        return [Syllable(**div, index=self.index + (index,))
                for index, div in enumerate(self.children)
                ]


@dataclass
class Division:
    """
    A class representing a division of a medieval chant document.
    """
    data: list  #: A list of data elements for the division.
    children: list  #: A list of child elements for the division.
    index: tuple

    def __post_init__(self):
        self.elements = []
        self.editorial_lines = []
        children_wo_paratext = self.filter_paratexts(self.children)
        if "data" in [keys for child in children_wo_paratext for keys in child.keys()]:

            self.interdivision = True
            self.elements = [Division(div["data"], div["children"], self.index + (index,)) for index, div in
                             enumerate(children_wo_paratext)]
            self.editorial_lines: list = self.get_flat_editorial_lines()
        else:
            self.interdivision = False
            self.elements = []
            self.editorial_lines: list = self.get_editorial_lines()  #: A list of `Syllable` objects representing the syllables in the division.

            # print("syllables", self.syllables)

        self.signature: str = self.get_signature()  #: The signature of the division, if present.
        self.status: str = self.get_status()  #: The status of the division, if present.

    def get_signature(self):
        division_metadata = {d["name"]: d["data"] for d in self.data}
        if "Signatur" in division_metadata.keys():
            return division_metadata["Signatur"]
        else:
            return None

    def get_status(self):
        dd = {d["name"]: d["data"] for d in self.data}
        if "Status" in dd.keys():
            return dd["Status"]
        else:
            return None

    def filter_paratexts(self, children):
        return [child for child in children if child["kind"] != "ParatextContainer"]

    def get_editorial_lines(self):
        editorial_lines = []
        for index, child in enumerate(self.filter_paratexts(self.children)):
            try:
                editorial_lines.append(EditorialLine(**child, index=self.index + (index,)))
            except:
                print("Warning: Expected editorial line but got: ", child)
        return editorial_lines

    def get_flat_editorial_lines(self):
        if len(self.elements):
            if self.elements[0].interdivision:
                return [editorial_line for division in self.elements for editorial_line in
                        division.get_flat_editorial_lines()]
            else:
                return [editorial_line for division in self.elements for editorial_line in division.editorial_lines]
        else:
            return []

    @property
    def flat_syllables(self):
        """ Get a list of syllables within a division. """
        return [syllable
                for editorial_line in self.editorial_lines
                for syllable in editorial_line.syllables]

    @property
    def flat_neumes(self):
        """ Get a list of neumes within a division. """
        return [neume
                for editorial_line in self.editorial_lines
                for syllable in editorial_line.syllables
                for neume in syllable.neumes]

    @property
    def flat_neume_components(self):
        """ Get a list of neume components within a division. """
        return [note_component
                for editorial_line in self.editorial_lines
                for syllable in editorial_line.syllables
                for neume in syllable.neumes
                for note_component in neume.neume_content]
        # if c["kind"] == "Syllable"


    @property
    def mei(self):
        if len(self.syllables) == 0:
            return ""
        syllables = "".join([syllable.mei for syllable in self.syllables])
        return f"<section><staff><layer>{syllables}</layer></staff></section>"

    @property
    def json(self):
        return {"type": "section", "elements": [syllable.json for syllable in self.syllables]}

    # def get_linechange(self):
    #     return [ Syllable(c) for child in self.children for d in child["children"] for c in d["children"] ]


class Chant:
    """
    A class representing a document or unit of medieval chant.

    A document is represented by its metadata and data,
    which are loaded from JSON files located in a directory specified
    by the entry parameter.
    """

    def __init__(self, entry_path):
        self.meta = self.get_meta(entry_path)
        data = self.get_data(entry_path)
        if data:
            self.data = data
        else:
            raise Exception("Could not load data", self.meta.id)
        self.type = self.meta.genre  #: The type of the document, taken from the gattung1 attribute of its metadata.

    @staticmethod
    def get_meta(entry_path):
        if glob.glob(entry_path + "/meta.json"):
            with open(entry_path + "/meta.json") as f:
                metadata = json.load(f)
                meta = Meta(**metadata)
            return meta
        else:
            return

    @staticmethod
    def get_data(entry):
        if glob.glob(entry + "/data.json"):
            with open(entry + "/data.json") as f:
                data = Data(**json.load(f))
            return data
        else:
            return None

    @property
    def flat_syllables(self):
        return [syllable
                for division in self.data.elements
                for el in division.editorial_lines
                for syllable in el.syllables]

    @property
    def flat_neumes(self):
        return [neume
                for division in self.data.elements
                for neume in division.flat_neumes]

    @property
    def flat_neume_components(self):
        try:
            return [neume_component
                    for division in self.data.elements
                    for neume_component in division.flat_neume_components]
        except:
            print("Warning for Chant property flat_neume_components: "
                  "data.elements is None at: ", self.meta.dokumenten_id)
            return []

    @property
    def volpiano(self):
        return [note_component.volpiano for note_component in self.flat_neume_components]

    @property
    def pitches(self):
        return [note_component.pitch for note_component in self.flat_neume_components]

    @property
    def flat_neume_components_by_division(self):
        return [[neume_component
                 for neume_component in division.flat_neume_components]
                for division in self.data.elements]

    @property
    def mei(self):
        try:
            return f'<?xml version="1.0" encoding="UTF-8"?>' \
                   '<?xml-model href="https://music-encoding.org/schema/dev/mei-Neumes.rng" ' \
                   'type="application/xml" schematypens="http://relaxng.org/ns/structure/1.0"?>' \
                   '<?xml-model href="https://music-encoding.org/schema/dev/mei-Neumes.rng" ' \
                   'type="application/xml" schematypens="http://purl.oclc.org/dsdl/schematron"?>' \
                   '<mei xmlns="http://www.music-encoding.org/ns/mei"><meiHead><fileDesc><titleStmt>' \
                   '<title></title></titleStmt><pubStmt></pubStmt></fileDesc></meiHead>' \
                   f'{self.data.mei}' \
                   '</mei>'
        except:
            print("Warning: Could not get MEI for: ", self.meta.dokumenten_id)
            return ""

    @property
    def json(self):
        return self.data.json


class Meta:

    def __init__(self, id, quelle_id, dokumenten_id, gattung1, gattung2, festtag, feier, textinitium,
                 bibliographischerverweis, druckausgabe, zeilenstart, foliostart, kommentar, editionsstatus,
                 additionalData):
        self.uuid = id,
        self.source_id = quelle_id
        self.document_id = dokumenten_id
        self.genre = gattung1
        self.subgenre = gattung2
        self.feast_day = festtag
        self.feast_time = feier
        self.initial_text = textinitium

        self.initial_folio = foliostart
        self.initial_line = zeilenstart
        self.ending_folio = additionalData.get("Endseite", "")
        self.ending_line = additionalData.get("Endzeile", "")

        self.bibliographical_reference = bibliographischerverweis
        self.cm_volume = druckausgabe

        self.editorial_comment = kommentar
        self.melody_number = additionalData.get("Melodiennummer_Katalog", "")
        self.melodyname_standardized = additionalData.get("Melodie_Standard", "")
        self.melodyname_diplomatic = additionalData.get("Melodie_Quelle", "")
        self.editor = additionalData.get("Editor", "")
        self.related_chant = additionalData.get("Bezugsgesang", "")
        self.liturgical_play_id = additionalData.get("Referenz_auf_Spiel", "")
        self.completeness = additionalData.get("Zusatz_zu_Textinitium", "")
        self.layer_of_addendum = additionalData.get("Nachtragsschicht", "")
        self.condition_of_transmission = additionalData.get("\u00dcberlieferungszustand", "")
        self.iiif_urls = additionalData.get("iiifs", "")

    @property
    def as_record(self):
        return {
            "uuid": self.uuid,
            "source_id": self.source_id,
            "document_id": self.document_id,
            "genre": self.genre,
            "subgenre": self.subgenre,
            "feast_day": self.feast_day,
            "feast_time": self.feast_time,
            "initial_text": self.initial_text,

            "initial_folio": self.initial_folio,
            "initial_line": self.initial_line,
            "ending_folio": self.ending_folio,
            "ending_line": self.ending_line,

            "bibliographical_reference": self.bibliographical_reference,
            "cm_volume": self.cm_volume,

            "editorial_comment": self.editorial_comment,
            "melody_number": self.melody_number,
            "melodyname_standardized": self.melodyname_standardized,
            "melodyname_diplomatic": self.melodyname_diplomatic,
            "editor": self.editor,
            "related_chant": self.related_chant,
            "liturgical_play_id": self.liturgical_play_id,
            "completeness": self.completeness,
            "layer_of_addendum": self.layer_of_addendum,
            "condition_of_transmission": self.condition_of_transmission,
            "iiif_urls": self.iiif_urls
        }


@dataclass
class Data:
    """
    A class representing data of a medieval chant document.
    """
    comments: list  # TODO: list of dicts
    uuid: str
    kind: str
    children: list  # = field(init=False) #: The Input Data that is parsed by the model
    documentType: str  #: The DocumentType attribute specifies the number of hierarchical levels
    # of Divisions the Document has
    elements: List[Division] = field(init=[])  #: The Division elements that are contained by the Document
    signatures: list = field(init=[])  #: A list of signatures for the elements of the document
    version: str = ""  #: to catch 'version' attribute in *some* data files

    def __post_init__(self):
        # try:
        self.elements = [Division(divisions["data"], divisions["children"], (index,)) for index, divisions in
                         enumerate(self.children) if "data" in divisions and "children" in divisions]
        self.signatures: List[str] = [e.signature for e in self.elements]

    # except:
    #    print("Element parsing did not work with doc: ", self.uuid)
    #    self.elements = []
    #    self.signatures = []

    @property
    def mei(self):
        try:
            divisions = "".join([division.mei for division in self.elements])
            return f'<music><body><mdiv><score><scoreDef />' \
                   f'{divisions}' \
                   f'</score></mdiv></body></music>'
        except AttributeError:
            print(f"Warning: Document {self.uuid} has no attribute 'elements'. Len of children attribute: {len(self.elements)}")
            print(self.elements)
            return ""

    @property
    def json(self):
        return {"type": "chant", "elements": [division.json for division in self.elements]}



from dataclasses import dataclass
from .document import Chant, Division
import re

class Song(Chant):
    pass

class Sequence(Chant):
    pass

class OrdinaryChant(Chant):
    pass

class PlayPassage(Chant):
    pass


from .corpus import Corpus
class Play(Corpus):
    def __init__(self, directory, liturgical_play_id, sample=None, filters=None):
        def cummulative_filter(document_meta, source_meta):
            if document_meta["liturgical_play_id"] != liturgical_play_id:
                return False
            if filters:
                return filters(document_meta, source_meta)
        super().__init__(directory, sample=sample, filters=cummulative_filter)

class ProperTropeComplex(Chant):
    """
    A class representing a trope complex.

    Parameters
    ----------
    entry : str
        The entry to initialize the `ProperTropeComplex` object.

    Attributes
    ----------
    nums : list
        List of signatures of trope elements
    letters : list
        List of signatures of prima chant elements
    ct_volume : str or None
        The Corpus Troporum Volume
    ct_base_chant : str or None
        ...

    Methods
    -------
    get_ct_base_chant()
        Extracts the base chant from the `meta.bibliographischerverweis` attribute using regular expressions.
    get_ct_volume()
        Extracts the volume from the `meta.bibliographischerverweis` attribute using regular expressions.
    get_nums() -> list
        Extracts numbers from the `meta.bibliographischerverweis` attribute and returns them as a list of integers.
    get_letters() -> list
        Extracts uppercase letters from the `meta.bibliographischerverweis` attribute and returns them as a list of strings.
    get_trope_elements() -> list
        Returns a list of divisions from the `data.elements` attribute of the `Document` whose status is "Tropenelement".
    get_primechant_elements() -> list
        Returns a list of divisions from the `data.elements` attribute of the `Document` whose status is "Einsatzmarke" (prime chant elements).
    """

    def __init__(self, entry):
        super().__init__(entry)
        self.nums = self.get_nums()
        self.letters = self.get_letters()
        self.ct_volume = self.get_ct_volume()
        self.ct_base_chant = self.get_ct_base_chant()

    def get_ct_base_chant(self):
        base_chant_match = re.search(r'^.*?(?=\s(?:\d|A))', self.meta.bibliographical_reference)
        return base_chant_match.group(0) if base_chant_match else None

    def get_ct_volume(self):
        volume_match = re.search(r'\(CT\s(\w+)\)$', self.meta.bibliographical_reference)
        return volume_match.group(1) if volume_match else None

    def get_nums(self) -> list:
        return [int(number) for number in re.findall(r"\d+", self.meta.bibliographical_reference)]

    def get_letters(self) -> list:
        return [letter for letter in self.meta.bibliographical_reference.split() if
                len(letter) == 1 and letter.isupper()]

    def get_trope_elements(self):
        return [division for division in self.data.elements if division.status == "Tropenelement"]

    def get_primechant_elements(self):
        return [division for division in self.data.elements if
                division.status == "Einsatzmarke"]  # Does not distinguish between complete and abbreviated (Einsatzmarke) prime chant elements

@dataclass
class TropeElement(Division):
    """
    A dataclass representing a trope element, inheriting from the `Division` class.
    """
    pass

@dataclass
class PrimeChantElement(Division):
    """
    A dataclass representing a prime chant element, inheriting from the `Division` class.
    """
    pass



from setuptools import find_packages, setup

with open("app/Readme.md", "r") as f:
    long_description = f.read()

setup(
    name="monodikit",
    version="0.0.1",
    description="",
    package_dir={"": "app"},
    packages=find_packages(where="app"),
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/timeipert/MonodiKit",
    author="Tim Eipert & Fabian C. Moss",
    author_email="tim.eipert@uni-wuerzburg.de",
    license="",
    classifiers=[""],
    install_requires=[],
    extra_require={
        "dev": ["twine>=4.0.2"]
    },
    python_requires=">=3.8",
)
